/*
 * cString.h
 *
 * cString that acts like std::string, but uses much smaller foot-print
 * and provides parsing capability
 */

#ifndef CSTRING_H_
#define CSTRING_H_


/**
 * Extended C++ String Class.
 *
 * cString s;
 * s = "Hello World...";
 * s.trimEnd(".");
 * printf("%s", s()); // cString becomes: "Hello World"
 *
 * s.erase("World");
 * printf("%s", s()); // cString becomes: "Hello "
 * s += "World";
 * printf("%s", s()); // cString back to: "Hello World"
 *
 * Use String class as an integer:
 * cString myInt;
 * myInt = 123;
 * myInt++; if(myInt == 124)    // true
 * myInt--; if(myInt == "123")  // true
 * int n = (int)myInt;          // Cast back from string to int, n = 123
 *
 */
class cString
{
    public:
        static int toInt(char* pString);   ///< Static method to convert char* string to integer
        static int toFloat(char* pString); ///< Static method to convert char* string to float

        cString();                      ///< Default constructor
        cString(int capacity);          ///< Constructor with initial capacity
        cString(const char* pString);   ///< Construct from char* pointer
        cString(const cString& s);      ///< Copy Constructor
        ~cString();                     ///< Destructor

        int getLen() const;   ///< @returns Number of characters in the string
        void clear();         ///< Clears the string

        void toLower(); ///< Make every alphabet character lowercase
        void toUpper(); ///< Make every alphabet character uppercase

        /**
         * @{ \name Insertion functions
         */
        void insertAtBeg(const char* pString);
        void insertAtBeg(const cString& s) { insertAtBeg(s()); }
        void insertAtEnd(const char* pString);
        void insertAtEnd(const cString& s) { insertAtEnd(s()); }
        void insertAt(const int index, const char* pString);
        void insertAt(const int index, const cString& s) { insertAt(index, s()); }
        /** @} */



        /**
         * @{ \name Append functions
         */
        void append(const char* pString);               ///< Appends constant string pointer
        void append(const cString& s) { append(s()); }  ///< Appends another cString
        void append(int x);                             ///< Appends integer as characters
        void append(float x);                           ///< Appends float as characters
        void appendAsHex(unsigned int num);             ///< Appends as hexadecimal ie: DEADBEEF
        /** @} */



        /**
         * @{ \name Number string to character string conversion functions
         * Example: Assign a number to a string: s=0xAA;
         * cString binary = s.toBinaryStr();
         * binary is now "10101010";
         */
        const char* toBaseNumberStr(const int base);
        const char* toBinaryStr() { return toBaseNumberStr(2);  }
        const char* toHexStr()    { return toBaseNumberStr(16); }
        /** @} */



        /**
         * @{ \name Comparison functions
         * @see See also the overloaded == and != operators
         */
        bool compareTo(const char* pString) const;
        bool compareTo(const cString& s) const;
        bool compareToIgnoreCase(const char* pString) const;
        bool compareToIgnoreCase(const cString& s) const { return compareToIgnoreCase(s()); }
        /** @} */



        /**
         * @{ \name Get Word Index Functions
         */
        int getWordIndexIgnoreCase(const char* pString) const;
        int getWordIndexIgnoreCase(const cString& s) const { return getWordIndexIgnoreCase(s()); }

        int firstIndexOf(const char* pString) const;
        int firstIndexOf(const cString& s) const { return firstIndexOf(s()); }
        int lastIndexOf(const char* pString) const;
        int lastIndexOf(const cString& s) const { return lastIndexOf(s()); }
        /** @} */



        /**
         * @{ \name Functions to check a contained string
         */
        bool contains(const char* pString) const;
        bool contains(const cString& s) const { return contains(s()); }
        bool containsIgnoreCase(const char* pString) const;
        bool containsIgnoreCase(const cString& s) const { return containsIgnoreCase(s()); }

        int countOf(const char* pString) const;
        int countOf(const cString& s) const { return countOf(s()); }

        bool endsWith(const char* pString) const;
        bool endsWith(const cString& s) const { return endsWith(s()); }
        bool endsWithIgnoreCase(const char* pString) const;
        /** @} */



        /**
         * @{ \name Erase Functions
         */
        bool erase(const char* pString);
        bool erase(const cString& s) { return erase(s()); }
        void eraseFirst(int nChars);
        void eraseLast(int nChars);
        void eraseCharAt(int index);
        void eraseAllAfter(int index);
        void eraseAfter(int index, int nChars);
        void eraseAllSpecialChars();
        /** @} */



        /**
         * @{ \name Trimming Functions to remove leading or trailing character sets
         */
        void trimStart(const char* pChars);
        void trimStart(const cString& s) { trimStart(s()); }
        void trimEnd(const char* pChars);
        void trimEnd(const cString& s) { trimEnd(s()); }
        /** @} */



        /**
         * @{ \name Find & Replace Functions
         */
        bool replaceFirst(const char* pFind, const char* pWith);
        bool replaceLast(const char* pFind, const char* pWith);
        int replaceAll(const char* pFind, const char* pWith);
        /** @} */



        /**
         * @{ \name Sub-string Functions
         */
        const char* subString(int fromIndex);
        const char* subString(int fromIndex, int charCount);
        const char* subString(char fromFirstChar);
        const char* subString(const char* fromStr);
        const char* subString(char fromFirstChar, int charCount);
        const char* subString(const char* fromStr, int charCount);
        const char* subString(char fromFirstChar, char toLastChar);
        /** @} */



        /**
         * @{ \name Tokenize functions
         * There are two ways to use tokenize operations:
         *  - One is to split all at once and get tokens by number
         *  - Second is to get tokens until an empty token is returned.
         * @note These tokenize operations perform manipulations into another memory region
         *          so this string itself maintains its contents.
         */
            /**
             * String can be split to tokens, and tokens can be obtained by index
             * @param pTokens contains all characters that are used to split into tokens
             *          ",'" means both , and ' can split the string, whichever is found first
             */
            int tokenizeAll(const char* pTokens);
            const char* getTokenNumber(int num);

            /**
             * This function begins token operation and returns the first token.
             * This method must be called before using getNextToken()
             * @param pToken  The tokens to get the string until
             *                  ie: If cString="hello,world", then getNextToken(",") will be "hello"
             */
            const char* beginTokenizeOperationGetFirstToken(const char* pToken);
            const char* getNextToken(const char* pToken); ///< @returns The next token given the token
        /** @} */



        /**
         * @{ \name Datatype detection Functions
         */
        bool isAllAlpha() const;     ///< @returns TRUE if the string is all alphabetical characters only
        bool isAlphaNumeric() const; ///< @returns TRUE if the string is all alpha or numerical characters only
        bool isFloat() const;        ///< @returns TRUE if the string is a floating point number
        bool isUint() const;         ///< @returns TRUE if the string is an unsigned number
        bool isInt() const;          ///< @retuns TRUE if the string is a number
        /** @} */


        /**
         * @{ \name Encryption & Checksum Functions
         */
        void encrypt(const char* pPassword);  ///< Encrypts (garbles) the string with given password
        void decrypt(const char* pPassword);  ///< Encrypted string must be decrypted before further use
        int getEncryptedStringLength() const; ///< Gets the number of characters that are currently encrypted

        unsigned int checksum_Get(); ///< Get integer value of XOR checksum of this string
        void checksum_Append();      ///< Appends checksum characters: Ex: 123 becomes: 123:0A
        void checksum_Remove();      ///< Removes checksum characters: Ex: 123:0A becomes 123
        bool checksum_Verify();      ///< @returns TRUE if for example: actual checksum 0A matches calculated 0A of string "123"
        /** @} */


        /**
         * @{ \name Assignment Operators
         */
        void operator=(const char* pString);    ///< Assign a string: myCStr = "123";
        void operator=(int num);                ///< Assign an int : myCStr = 123;
        void operator=(float num);              ///< Assign a float: myCStr = 1.23;
        cString& operator=(const cString& rhs); ///< Assign Operator for cString a = cString b
        /** @} */

        /**
         * @{ \name Increment & Decrement Operators if the string consist of integer value
         */
        void operator++();           ///< Pre-Increment if String is an integer
        void operator++(int unused); ///< Post-Increment if String is an integer
        void operator--();           ///< Pre-Decrement if String is an integer
        void operator--(int unused); ///< Post-Decrement if String is an integer
        /** @} */


        /**
         * Add/Subtract Operators:
         */
        void operator+=(const char singleChar); ///< Append char using += Operator
        void operator+=(const char* pString);   ///< Append String using += Operator
        void operator+=(int n);                 ///< Subtract Integer using += Operator if string is an integer value
        void operator+=(float n);               ///< Subtract Float using += Operator if string is an float value
        void operator-=(const char* pString);   ///< Remove all instances of pString using -= Operator
        void operator-=(int n);                 ///< Subtract Integer String using if string is an integer value
        void operator-=(float n);               ///< Subtract Float String using -= Operator if string is an float value
        /** @} */


        /**
         * @{ \name Equality Operators
         */
        bool operator==(const char* pString) const;
        bool operator==(const cString& rhs) const;
        bool operator==(int) const;
        bool operator!=(const char* pString) const;
        bool operator!=(int) const;
        // Note: Floats cannot be compared using == operator, they require delta comparison
        /** @} */

        /**
         * @{ \name Comparison Operators to perform comparisons, such as:
         * cString s;
         * s = "hello";  if(s < "helloo") ...
         * s = 123;      if(s < 1234) ...
         * s = 1.23;     if(s >= 1.22) ...
         */
        bool operator<(const char* pString) const;
        bool operator<(int) const;
        bool operator<(float) const;
        bool operator>(const char* pString) const;
        bool operator>(int) const;
        bool operator>(float) const;
        bool operator<=(const char* pString) const;
        bool operator<=(int) const;
        bool operator<=(float) const;
        bool operator>=(const char* pString) const;
        bool operator>=(int) const;
        bool operator>=(float) const;
        /** @} */

        const char* operator()() const;                   ///< () Operator: Ex: puts(myCStr());
        const char* c_str() const { return (*this)(); };  ///< Get c-string pointer (calls operator above)

        operator int() const;       ///< (int) Cast Operator: Ex: int x = (int)myCStr;
        operator float() const;     ///< (float) Cast Operator: Ex: float x = (float)myCStr;
        char& operator[](int pos);  ///< Index Operator to get and set value @ Index

    private:
        /// init() is called by constructor to initialize the string
        void init(int initialLength=16);

        /// Ensures that the string contains enough memory to store additional nChars characters
        void ensureMemoryToInsertNChars(const int nChars);

        /// Ensures that the backup string used to tokenize and perform substring operations has enough memory for nChars
        void ensureMemoryForBackupString(int nChars);

        /// reallocates memory for this string given the new size
        void reAllocateMem(const int size);

        /// copies to our string the string given by pString
        void copyFrom(const char* pString);

        /// Converts single hexadecimal to int, such as "A" to the value of 10
        int singleHexCharToInt(unsigned char theChar);

        /// Converts hexadecimal value of string to its integer equivalent (base 10)
        int hexStrDigitsToInt(char* pString);

        /// Encryps the string given the password and the length to encrypt
        void encrypt(const char* pPassword, const int lenToEncrypt);

        int mCapacity;              ///< Capacity of the memory of this string
        int mLenOfEncryptedChars;   ///< Length of characters that were originally encrypted
        int mBackupStrCapacity;     ///< The capacity of the backup string used for substring() and tokenize() functions
        char* mpStr;                ///< Pointer to the primary memory
        char* mpBackupStr;          ///< Pointer to the backup memory used for substring() and tokenize()
        char* mpTokenPtr;           ///< The pointer to the token
        static const int mInvalidIndex = -1;
};

#endif /* CSTRING_H_ */
